__doc__ = """
HDF5-based file format for heterogeneous numerical data.
Based on code from and inspired by

- HEPfile: https://github.com/mattbellis/hepfile
- NeuroH5: https://github.com/iraikov/neuroh5

.. automodule:: miv.io.file.read
   :members:

.. automodule:: miv.io.file.write
   :members:

"""

from miv.io.file.read import *
from miv.io.file.write import *
