__doc__ = """
Sample datasets for debugging/testing.

Any downloaded datasets will be saved in **datasets** directory.
If the dataset already exists, download will be skipped.
To remove any cache files, you can delete the **datasets** directory.
"""

from miv.datasets.optogenetic import load_data
