__doc__ = """
Auditory Model
==============

References
----------

- https://engineering.purdue.edu/~malcolm/interval/1998-010/
"""

from miv.coding.temporal.lyon import *
from miv.coding.temporal.protocol import *
from miv.coding.temporal.spiker import *
